#include <VRS51L3074_SDCC.h>

#include <stdio.h>
#include <string.h>
#include <ctype.h>

extern int atoi(char *);

#include "../HD61202/lowlevel.h"
#include "../HD61202/highlevel.h"
#include "../HD61202/highlevel_text.h"

#include "../ProcComp/i2c.h"
#include "../ProcComp/RTC.h"

#include "../serials/serial0.h"
#include "../serials/serial1.h"

#include "../softtimers/softtimers.h"



#define HOLD_TIME	5
#define BEEP_TIME	10

// some parts of the code do not support more than 99 entries (list), others handle ints
#define NUM_ENTRIES	60

#define NUM_RECORDS	256


// check display width, becasue print routine does not...
#define NAME_LEN 16

struct RTCtm
{
  unsigned char tm_sec;                   /* Seconds.     [0-60]  BCD   */
  unsigned char tm_min;                   /* Minutes.     [0-59]  BCD   */
  unsigned char tm_hour;                  /* Hours.       [0-23]  BCD   */
  unsigned char tm_wday;                  /* Day of week. [0-6]         */
  unsigned char tm_mday;                  /* Day.         [1-31]  BCD   */
  unsigned char tm_mon;                   /* Month.       [0-11]  BCD   */
  unsigned char tm_year;                  /* Year since 2000      BCD   */
};

typedef struct {
   char name[NAME_LEN+1];
   unsigned char rfid[6];
} rfid_record;

typedef struct {
   unsigned int id;
   struct RTCtm time;
} time_record;

struct ptrstruct {
	unsigned int index;
	unsigned int count;
	unsigned int bkindex;
	unsigned int bkcount;
	unsigned char tag;
};

__xdata at 0x8000 rfid_record rfid_base[NUM_ENTRIES];
__xdata at 0x8000+sizeof(rfid_record)*NUM_ENTRIES time_record time_base[NUM_RECORDS];
__xdata at 0x9F00 struct ptrstruct bufpos;
__xdata at 0x9F00+sizeof( struct ptrstruct) struct ptrstruct bufpos2;
__xdata char linebuf[21],serbuf[21+16+5],learnbuf[21];

unsigned int entries,rcrdidx,rcrdcnt;

__bit  redraw,compact;

__xdata struct RTCtm rtcval;

/* Program states */
#define ESCRACHATE	1
#define LEARN		2
static int sstate;

unsigned char bcd2bin(unsigned char bcd)
{
unsigned char aux;

   aux=bcd>>4;
   return(bcd-6*aux);
}

unsigned char bin2bcd(unsigned char bin)
{
unsigned char aux;

	aux=bin/10;
   return(bin+6*aux);
}


void printdatetime(__xdata struct RTCtm *rtc)
{
__code const static char *  __code months[]={
"---","ENE","FEB","MAR","ABR","MAY","JUN","JUL","AGO","SEP","OCT","NOV","DIC"
};
__xdata char *ptr;
__code char *cptr;
register unsigned char aux;

	ptr=linebuf;
	cptr=months[bcd2bin(rtc->tm_mon)];
	aux=rtc->tm_mday;
	*(ptr++)=(aux>>4)+0x30;
	*(ptr++)=(aux&0x0F)+0x30;
	*(ptr++)=' ';
	*(ptr++)=*(cptr++);
	*(ptr++)=*(cptr++);
	*(ptr++)=*(cptr++);
	*(ptr++)=' ';
	aux=rtc->tm_year;
	*(ptr++)=(aux>>4)+0x30;
	*(ptr++)=(aux&0x0F)+0x30;
	*(ptr++)=' ';
	aux=rtc->tm_hour;
	*(ptr++)=(aux>>4)+0x30;
	*(ptr++)=(aux&0x0F)+0x30;
	*(ptr++)=':';
	aux=rtc->tm_min;
	*(ptr++)=(aux>>4)+0x30;
	*(ptr++)=(aux&0x0F)+0x30;
	*(ptr++)=':';
	aux=rtc->tm_sec;
	*(ptr++)=(aux>>4)+0x30;
	*(ptr++)=(aux&0x0F)+0x30;
	*(ptr++)=0;
}


int datetime(void)
{
  unsigned char len;
  
       if(ser0txbytes)
             return(0);
       printdatetime(&rtcval);
       strcpy(serbuf,linebuf);
       len=strlen(serbuf);
       serbuf[len++]='\r';
       serbuf[len++]='\n';
       ser0send(serbuf,len);
       return(-1);
}


int list(void)
{
static unsigned int printline=0;
unsigned char len;

       if(ser0txbytes)
             return(printline);
       if(printline < entries){
	  len=bin2bcd(printline+1);
	  serbuf[0]=(len>>4)+0x30;
	  serbuf[1]=(len&0x0F)+0x30;
	  serbuf[2]='\t';
	  strcpy(&serbuf[3],rfid_base[printline].name);
	  len=strlen(serbuf);
	  serbuf[len++]='\r';
	  serbuf[len++]='\n';
	  ser0send(serbuf,len);
	  return(++printline);	// Increment line number
	}
        else {
	  printline=0;
          return(-1);
	}
}


void learn()
{
unsigned char len;

	if(entries<NUM_ENTRIES){
	  len=(ser0rxbytes<=NAME_LEN+2)?ser0rxbytes-2:NAME_LEN;
	  ser0rxbuf[len+1]=0;
	  strcpy(learnbuf,&ser0rxbuf[2]);
	  sstate=LEARN;
	  redraw=1;
	} else {
	  // FULL
	}
}

void delete()
{
unsigned int id,i;

       id=atoi(&ser0rxbuf[2]);
       if((id<=entries)&&entries&&id){
	 id--;
	 for(i=id+1;i<entries;i++){
	   strcpy(rfid_base[i-1].name,rfid_base[i].name);
	   memcpy(rfid_base[i-1].rfid,rfid_base[i].rfid,5);
	 }
	 entries--;
	 updateFRAM2();
       }
//  WARNING NOT POWEROFF-FRIENDLY !!!
}

void updateFRAM(void)
{
    	bufpos.bkindex=rcrdidx;         // save to backup copy
	bufpos.bkcount=rcrdcnt;
	bufpos.tag=0xFF;                       // invalidate main copy
	bufpos.index=rcrdidx;	// save to main copy
	bufpos.count=rcrdcnt;
	bufpos.tag=0;                     // validate main copy
}

void updateFRAM2(void)
{
    	bufpos2.bkcount=entries;         // save to backup copy
	bufpos2.tag=0xFF;                       // invalidate main copy
	bufpos2.count=entries;	// save to main copy
	bufpos2.tag=0;                     // validate main copy
}

void purgeid()
{
	entries=0;
	updateFRAM2();
}

void purgedb(void)
{
	rcrdidx=0;
	rcrdcnt=0;
	updateFRAM();
	redraw=1;
}


void setdatetime()
{
unsigned int month, day, year, hour, minute;


    day = atoi(&ser0rxbuf[2]);
    if (day > 0 && day < 32) {
      month = atoi(&ser0rxbuf[5]);
      if (month > 0 && month < 32) {
	year = atoi(&ser0rxbuf[8]);
	if (year < 100) {
	  hour = atoi(&ser0rxbuf[11]);
	  if (hour < 24) {
	    minute = atoi(&ser0rxbuf[14]);
	    if (minute < 60) {
	      rtcval.tm_sec  = 0;	   		// 0-59
	      rtcval.tm_min  = bin2bcd(minute);		// 0-59
	      rtcval.tm_hour = bin2bcd(hour);		// 0-23
	      rtcval.tm_mday = bin2bcd(day);		// 1-31
	      rtcval.tm_mon  = bin2bcd(month);		// 1-12
	      rtcval.tm_year = bin2bcd(year);		// 0-99
	      RTCSet(0xD0, &rtcval);
	    }
	  }
	}
      }
    }
}


void decode(__xdata unsigned char *rfid)
{
/*int i;
  for(i=0;i<5;i++){
    rfid[2*i+1]-=0x30;
    if(rfid[2*i+1]>=10)
      rfid[2*i+1]-=7;
    rfid[2*i+2]-=0x30;
    if(rfid[2*i+2]>=10)
      rfid[2*i+2]-=7;
    rfid[i]=(rfid[2*i+1]<<4) | rfid[2*i+2];
  }
*/
__xdata unsigned char *ptr;
unsigned char i;

  ptr=rfid+1;
  i=5;
  while(i--){
    *ptr-=0x30;
    if(*ptr>=10)
      *ptr-=7;
    ptr++;
    *ptr-=0x30;
    if(*ptr>=10)
      *ptr-=7;
    *(rfid++)=(*(ptr-1)<<4) | *ptr;
    ptr++;
  }
}



int lookup(unsigned char *id)
{
int i;
	for(i=0;i<entries;i++){
   	if(!memcmp(id,rfid_base[i].rfid,5))
      	return(i);
   }
   return(-1);
}


void showonscreen(void)
{
register int record;
register int j;


	record=0;
	if(compact){
	  while((record<5)&&(rcrdcnt>record)){
		j=rcrdidx-record-1;
		if(j<0)
		  j+=NUM_RECORDS;
		printdatetime(&time_base[j].time);
		memcpy(linebuf,linebuf+10,5);          // extract hour:min
		linebuf[5]='|';
		memcpy(&linebuf[6],rfid_base[time_base[j].id].name,14);
		linebuf[20]=0;
		LCD_print6at(3+record++,0,linebuf);
	  }
	}
        else {
	  while((record<2)&&(rcrdcnt>record)){
		j=rcrdidx-record-1;
		if(j<0)
		  j+=NUM_RECORDS;
		LCD_print8at(3+(record<<1),0,0,rfid_base[time_base[j].id].name);
		printdatetime(&time_base[j].time);
		LCD_print6at(4+(record<<1),1,linebuf);
		record++;
	  }
	}
}

int show(void)
{
static int printline=0;
static unsigned char len;
register int j;
__xdata char *name;

       if(ser0txbytes)
             return(printline);
       if(printline < rcrdcnt){
	  j=rcrdidx-rcrdcnt+printline;
	  if(j<0)
	    j+=NUM_RECORDS;
	  else if(j>=NUM_RECORDS)
	    j-=NUM_RECORDS;
	  printdatetime(&time_base[j].time);
	  strcpy(serbuf,linebuf);
	  strcat(serbuf,"\t");
	  name=rfid_base[time_base[j].id].name;
	  strcat(serbuf,name);
	  len=strlen(serbuf);	  
	  serbuf[len++]='\r';
	  serbuf[len++]='\n';
	  ser0send(serbuf,len);
	  return(++printline);	// Increment line number
	}
        else {
	  printline=0;
          return(-1);
	}
}

const static __code unsigned char rtcstart[]={0,0,0,0,1,1,7};
        

/* MAIN PROGRAM */


main()
{
int i;
__xdata unsigned char RFID[15];
__code const static char wrong[]="El RFID leido";
__bit  listing=0,showing=0,datetiming=0,rxing0=0,rxing1=0;

// Allocate software timers
#define timer SEC_TIMERS[0]
#define r0timer SEC_TIMERS[1]
#define btimer CS_TIMERS[0]
#define r1timer CS_TIMERS[1]
#define dbnctmr CS_TIMERS[2]

#define SWITCH0 (!P3_2)


  DEVMEMCFG  |= 0xC0;					// Activate FRAM
  init_softtimers();
  sstate=ESCRACHATE;
  //	bufpos.index=0;                                 // First init
  //	bufpos.count=0;
  //    bufpos.tag=0;
  //    bufpos2.count=0;
  //    bufpos2.tag=0;
  if(bufpos.tag){                                       // crash ?
    rcrdidx=bufpos.bkindex;                             // Recover pointers from backup set in FRAM
    rcrdcnt=bufpos.bkcount;
  }
  else {
    rcrdidx=bufpos.index;                                 // Recover pointers from FRAM
    rcrdcnt=bufpos.count;
  }
  if(bufpos2.tag){                                       // crash ?
    entries=bufpos2.bkcount;                             // Recover pointers from backup set in FRAM
  }
  else {
    entries=bufpos2.count;                                 // Recover pointers from FRAM
  }

	
  redraw=1;
  compact=1;
  LCD_init();
  LCD_clear();
  LCD_print8at(1,1,'B',"Time logger R2");
  LCD_print8at(3,3,'B',"v1.0 READY");
  LCD_rectangle(4,4,123,44);
  LCD_print6at(7,1,"* Cika Electronica");
  Init_I2C();
  Init_RTC();
  // Start oscillator, we don't have a battery/capacitor (OSCEN=0 in CAL/Control register) 
  RTCWrite(0xD0,0x01,0x00);
  // This is a demo, so setup the RTC to start at 0
  RTCSet(0xD0, rtcstart);
  ser0init(40000000/9600);
  P5PINCFG&=~0x1F; // P5.0,1,2,3,4 output
  P5=0;
  P4PINCFG&=~(1<<4); // P4.4 output
  P3PINCFG|=((1<<3)|(1<<2)); // P3.2,3 input
  ser1init(40000000/9600);
  GENINTEN = 0x03; //Enable global interrupts
  timer=HOLD_TIME;
  while(1){

        if(!btimer)
	  P5&=~0x03;                         // Buzzers/voice off
	
	if(listing)
	  if(list()==-1)
	    listing=0;
	if(showing)
	  if(show()==-1)
	    showing=0;
	if(datetiming)
	  if(datetime()==-1)
	    datetiming=0;

	RTCReadAll(0xD0,(unsigned char *)&rtcval);		//Read Current RTC values
        if(!timer){ // do not update screen if timer is set
	  if(redraw){// do not clear screen if not necessary
     	  	LCD_clear();
		LCD_print8at(0,0,0,"Cika Time Logger");
		LCD_line(0,17,127,17);
		if(sstate==LEARN)
		  LCD_print8at(7,0,0,"*Acerque el RFID");
		else 
		  showonscreen();
		redraw=0;
	  }
	  printdatetime(&rtcval);
	  LCD_print6at(1,1,linebuf);
	}
	else redraw=1;
	// Handle commands
	if(ser0rxbytes&&!rxing0){         // some data in buffer
	  r0timer=30;                     // start RFID timeout timer
	  rxing0=1;
	}
	else if(rxing0&&!r0timer){        // timer expired and packet incomplete
	  ser0flush();                    // discard incomplete data
	  rxing0=0;
	}
	if((ser0peek()==0x0D) && !listing && !showing && !datetiming){
	  switch(toupper(ser0rxbuf[0])){
	  case 'L':
	    learn();
	    break;
	  case 'I':
	    listing=1;
	    break;
	  case 'R':
	    showing=1;
	    break;
	  case 'P': 
	    switch(toupper(ser0rxbuf[1])){
	    case 'R':
	      purgedb();
	      break;
	    case 'I':
	      purgeid();
	      break;
	    }
	    break;
	  case 'T':
	    datetiming=1;
	    break;
	  case 'D':
	    delete();
	    break;
	  case 'S':
	    if(toupper(ser0rxbuf[1])=='T'){
	      setdatetime();
	      datetiming=1;
	    }
	    break;
	  }
	  ser0flush();
	  rxing0=0;
	}
	// Handle mode switch
	if(SWITCH0){
	  if(!dbnctmr){
	    compact^=1;
	    redraw=1;
	  }
	  dbnctmr=6;
	}
	// Handle RFID record
	if(ser1rxbytes&&!rxing1){         // some data in buffer
	  r1timer=30;                     // start RFID timeout timer
	  rxing1=1;
	}
	else if(rxing1&&!r1timer){        // timer expired and packet incomplete
	  ser1flush();                    // discard incomplete data
	  rxing1=0;
	}
	if(ser1rxbytes==14){
	        memcpy(RFID,ser1rxbuf,14);
		ser1flush();
		rxing1=0;                // no timeout
         	if((RFID[0]==2)&&(RFID[13]==3)){
		  decode(RFID);
	      	  LCD_clear();
		  switch (sstate){
		  case ESCRACHATE:
		    if((i=lookup(RFID))!= -1){
		      printdatetime(&rtcval);      // rtcval will change when we read from RTC
		      LCD_print6at(2,1,linebuf);
		      LCD_print8at(4,0,0,rfid_base[i].name);
		      time_base[rcrdidx].id=i;
		      memcpy((void *) &time_base[rcrdidx].time,(void *) &rtcval,sizeof(struct RTCtm));
		      rcrdcnt++;			// inc record count
		      if(++rcrdidx>=NUM_RECORDS)	// keep track of circular boundaries for index
			rcrdidx=0;
		      if(rcrdcnt >= NUM_RECORDS)	// if storage is full, account for overwritten rcrdidx
			rcrdcnt=NUM_RECORDS;		
		      // update index and record count in FRAM
		      updateFRAM();
		      P5_0=1;
		    }
		    else {
		      LCD_print6at(2,3,wrong);
		      LCD_print6at(4,0,"no figura en la base");
		      P5_1=1;
		    }
		    btimer=BEEP_TIME;
		    break;
		  case LEARN:
		    if((i=lookup(RFID))!= -1){
		      LCD_print6at(2,3,wrong);
		      LCD_print6at(4,0,"ya figura en la base");
		    }
		    else {
		      LCD_print6at(0,0,"Ingresado como:");
		      strcpy(rfid_base[entries].name,learnbuf);
		      memcpy(rfid_base[entries].rfid,RFID,5);
		      LCD_print8at(2,0,0,rfid_base[entries++].name);
		      // update index and record count in FRAM
		      updateFRAM2();
		    }
		    sstate=ESCRACHATE;
		    break;
		  default:
		    break;
		  }
		  timer=HOLD_TIME;					// restart hold timer
		}
	 }
  }
}
